.pnx-resource a {
  color: var(--accent-foreground-rest);
  text-decoration: none;
}
.pnx-resource a:hover {
  color: var(--accent-foreground-hover);
}
.pnx-resource a svg {
  color: var(--neutral-foreground-rest);
}
.pnx-resource a svg:hover {
  color: var(--neutral-foreground-hover);
}
